/*
 * This file is made available under the terms of the LGPL licence.
 * This licence can be retreived from http://www.gnu.org/copyleft/lesser.html.
 * The source remains the property of the YAWL Group.  The YAWL Group is a collaboration of
 * individuals and organisations who are commited to improving workflow technology.
 *
 */

/**
 * To develop your own service inherit from this class and implement the method doMyService().
 */

package service;

import org.yawlfoundation.yawl.elements.data.*;
import org.yawlfoundation.yawl.engine.YSpecificationID;
import org.yawlfoundation.yawl.engine.interfce.*;
import org.yawlfoundation.yawl.engine.interfce.interfaceB.*;
import org.yawlfoundation.yawl.engine.interfce.interfaceA.*;

import org.jdom.Element;

import java.util.*;
import java.io.*;

import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import org.yawlfoundation.yawl.authentication.User;

import org.jdom.output.XMLOutputter;

import yawlservice.*;

public abstract class AbstractYawlService
    extends InterfaceBWebsideController {

  private static final String ENGINE_URI = "http://localhost:8080/yawl/ia";
  private static final String DECLARE_SERVICE_URI = "http://localhost:8080/declareService/ib";
  
  protected static final String ADMIN_USERNAME = DEFAULT_ENGINE_USERNAME;
  protected static final String ADMIN_PASSWORD = DEFAULT_ENGINE_PASSWORD;
  
  // required data for interfacing with the engine
  private String DECLARE_USERNAME = "declareService" ;
  private String DECLARE_PASSWORD = "declare" ;

  private InterfaceA_EnvironmentBasedClient interfaceAClient;
  
  //
  private String sessionHandle = null;
  private Log log;

  /**
   *
   */
  public AbstractYawlService(String logFile) {
    super();
    log = new Log(logFile);
    interfaceAClient = new InterfaceA_EnvironmentBasedClient(ENGINE_URI);
    log.log("started"); // write to log
  }
  
  protected Log getLog(){
	  return log;
  }
  
  //***************************************************************************//

  /** Checks if there is a connection to the engine, and
   *  if there isn't, attempts to connect
   *  @return true if connected to the engine
   */
  private boolean connected() {
      try {
          // if not connected
           if ((sessionHandle == null) || (!checkConnection(sessionHandle)))
              sessionHandle = connectAsService();
      }
      catch (IOException ioe) {
          log.log("Exception attempting to connect to engine: ");
          log.exception(ioe);
      }
      return (successful(sessionHandle)) ;
  }
  
  private String connectAsService() throws IOException{

      // first connect as default admin user
      sessionHandle = connect(ADMIN_USERNAME, ADMIN_PASSWORD);

      // create new user for service if necessary
      if (! isRegisteredUser(DECLARE_USERNAME))
         interfaceAClient.createUser(DECLARE_USERNAME, DECLARE_PASSWORD, true, sessionHandle);

      // logon with service user and return the result
      return connect(DECLARE_USERNAME, DECLARE_PASSWORD);
  }

  /**
   * Checks if a user is currently registered for this session
   * @param user
   * @return true if the user is registered in the current session
   */
  private boolean isRegisteredUser(String user) {

      // check if service is a registered user
      try {
          ArrayList<User> users = (ArrayList<User>) interfaceAClient.getUsers(sessionHandle);
          Iterator<User> itr = users.iterator();
          while (itr.hasNext()) {
              User u = (User) itr.next() ;
              if ( u.getUserID().equals(user) ) return true ;       // user in list
          }
          return false;                                             // user not in list
      }
      catch (IOException ioe) {
          return false;                                             // no list 
      }

  }
  
  /**
   * This method is called after the work item was checkedOut. After this method the work item will be checkedIn.
   *
   * @param external ExternalWorkItem
   */
  protected abstract void doMyService(ExternalWorkItem external);

  /**
   *
   * @param workItemRecord WorkItemRecord
   */
  public void handleEnabledWorkItemEvent(WorkItemRecord workItemRecord) {
    try {
      if (connected()) {
        if ( checkOut(workItemRecord.getID(), sessionHandle) == null ) {
        }
        else {
          List<WorkItemRecord> children = super.getChildren(workItemRecord.getID(),
                                            sessionHandle);
          // actually, you have to check back IN a child of the work item.
          // we do that by copying the engineID of the first child into the work item, and then checing
          // the work item IN
          if (children.size() > 0) {
            WorkItemRecord child = (WorkItemRecord) children.get(0);
            ExternalWorkItem external = this.getExternalWorkItemAndData(child);
            this.doMyService(external); // execute whatever you want with the work item
          }
        }
      }
    }
    catch (Exception e) {
      log.exception(e);
    }
  }

  public String getStackTrace(Throwable t) {
    StringWriter stringWritter = new StringWriter();
    PrintWriter printWritter = new PrintWriter(stringWritter, true);
    t.printStackTrace(printWritter);
    printWritter.flush();
    stringWritter.flush();
    return stringWritter.toString();
  }

  /**
   * Given the work item record get all information and data elements.
   * @param workItemRecord WorkItemRecord
   * @return ExternalWorkItem
   */
  private ExternalWorkItem getExternalWorkItemAndData(WorkItemRecord
      workItemRecord) {
    try {
      //*** create external work item: taskID, caseId, specificationID, decompositionID...
      String taskID = workItemRecord.getTaskID();
      String caseID = workItemRecord.getCaseID();
      String specificationID = workItemRecord.getSpecificationID();
      TaskInformation taskinfo = getTaskInformation(new YSpecificationID(specificationID), taskID,
          sessionHandle);
      String decompositionID = taskinfo.getDecompositionID();
      String engineID = workItemRecord.getID();
      ExternalWorkItem external = new ExternalWorkItem(taskID, caseID,
          specificationID, decompositionID, engineID);

      // *** read input and output data for the external work item
      getOutputData(external, taskinfo);
      getInputData(workItemRecord, external, taskinfo);

      // *** check out all children of the work item
      List<WorkItemRecord> children = super.getChildren(workItemRecord.getID(),
                                        sessionHandle);
      for (int i = 0; i < children.size(); i++) {
        WorkItemRecord itemRecord = (WorkItemRecord) children.get(i);
        if (WorkItemRecord.statusFired.equals(itemRecord.getStatus())) {
          checkOut(itemRecord.getID(), sessionHandle);
        }
      }
      return external;
    }
    catch (Exception e) {
      log.exception(e);
      return null;
    }
  }

  /**
   *
   * @param external ExternalWorkItem
   * @param sessionHandle String
   * @return String
   */
  public synchronized void checkInWorkItem(ExternalWorkItem
                                           external) {
    try {
      if (connected()) {
        Element task = prepareReplyRootElement(external, sessionHandle);

        Element outputData = this.outputData(external, task);
        Element inputData = this.inputData(external, task);

         
        String result = checkInWorkItem(external.getEngineID(), inputData, outputData,
                        sessionHandle);
        if (successful(result)){
        	log.logEvent("CHECKED IN", external); // log this event
        
        } else {
        	log.logEvent("FAILED TO CHECK IN", external); // log this event
        }     
      }
    }
    catch (Exception e) {
      log.exception(e);
    }
  }

  /**
   * This is something that I saw somewhere in YAWL code. To be honoust, I don't
   * exactly know what this means.
   *
   * @param external ExternalWorkItem
   * @param sessionHandle String
   * @return Element
   * @throws IOException
   */
  protected Element prepareReplyRootElement(ExternalWorkItem external,
                                            String sessionHandle) throws
      IOException {
    Element replyToEngineRootDataElement;

    //prepare reply root element.
    SpecificationData sdata = getSpecificationData(new YSpecificationID(
        external.getSpecificationID()),
        sessionHandle);

    String decompID = external.getDecompositionID();
    if (sdata.usesSimpleRootData()) {
      replyToEngineRootDataElement = new Element("data");
    }
    else {
      replyToEngineRootDataElement = new Element(decompID);
    }
    return replyToEngineRootDataElement;
  }

  /**
   * For the extrenal work item we get all output data from the YAWL engine.
   *
   * @param external ExternalWorkItem
   * @param taskInfo TaskInformation
   */
  private void getOutputData(ExternalWorkItem external,
                             TaskInformation taskInfo) {
    YParametersSchema params;
    if (taskInfo == null) {
      return;
    }
    params = taskInfo.getParamSchema();
    List<YParameter> _output = params.getOutputParams();
    YParameter parameter = null;     
    for (int i = 0; i < _output.size(); i++) {
      parameter = (YParameter) _output.get(i);
      String value = parameter.getInitialValue();
      external.addOutputData(parameter.getName(),parameter.getDataTypeName(), value);
    }
  }

  /**
   * For the extrenal work item we get all input data from the YAWL engine.
   *
   * @param external ExternalWorkItem
   * @param taskInfo TaskInformation
   */
  private void getInputData(WorkItemRecord workItemRecord,
                            ExternalWorkItem external,
                            TaskInformation taskInfo) {
    org.jdom.Element inputData = workItemRecord.getDataList();
    YParametersSchema params;
    if (taskInfo == null) {
      return;
    }
    params = taskInfo.getParamSchema();
    List<YParameter> _output = params.getInputParams();
    YParameter parameter = null; ;
    for (int i = 0; i < _output.size(); i++) {
      parameter = (YParameter) _output.get(i);

      String name = parameter.getName(); // name

      String value = parameter.getInitialValue(); // value
      if (inputData != null) {
        value = ( (Element) inputData.getChild(name)).getText();
      }

      String type = parameter.getDataTypeName(); // type

      external.addInputData(name,type, value);
    }
  }

  /**
   *
   * @param workItemRecord WorkItemRecord
   */
  public void handleCancelledWorkItemEvent(WorkItemRecord workItemRecord) {
    // add code here if necessary
  }

  /**
   * Just creating XML form output data. We need this method to write the extrenal work item
   * to the log fine. We use xml format for the data.
   *
   * @param external ExternalWorkItem
   * @param root Element
   * @return Element
   */
  private Element outputData(ExternalWorkItem external, Element root) {
    ExternalDataElement data = null;
    Element dataXML = null;
   for (int i = 0; i < external.dataSize(); i++) {
        data = (ExternalDataElement) external.getData(i);
        if (data.isOutput()){
          dataXML = new Element(data.getName());
          dataXML.setText(data.getValue());
          root.addContent(dataXML);
        }
    }
    return root;
  }

  /**
   * Just creating XML form input data. We need this method to write the extrenal work item
   * to the log fine. We use xml format for the data.
   *
   * @param external ExternalWorkItem
   * @param root Element
   * @return Element
   */
  private Element inputData(ExternalWorkItem external, Element root) {
    ExternalDataElement data = null;
    Element dataXML = null;
   for (int i = 0; i < external.dataSize(); i++) {
     data = (ExternalDataElement) external.getData(i);
     if (data.isInput()) {
       dataXML = new Element(data.getName());
       dataXML.setText(data.getValue());
       root.addContent(dataXML);
     }
   }
    return root;
  }

//***************************************************************************//
  /**
   *  Starts a worklet case executing in the engine
   *  @param coci - the item record of the child workitem to start the
   *                worklet for
   *  @return - the case id of the started worklet case
   */
  protected String launchExternalCase(ExternalCase extCase) {
    if (uploadSpecification(extCase)) {
      String caseId;
      String specID = extCase.getStecificationID();

      // fill the case params with matching data values from the workitem
      String caseData = mapItemParamsToWorkletCaseParams(extCase);

      try {
        caseId = _interfaceBClient.launchCase(new YSpecificationID(specID), caseData, sessionHandle,DECLARE_SERVICE_URI);
          if (successful(caseId)) {
          extCase.setCaseID(caseId);
          return caseId;
        }
        else {
          log.log("Unable to launch case for: " + specID);
          log.log("Diagnostic message: " + caseId);
          return null;
        }
      } catch (IOException ioe) {
        log.exception(ioe);
        return null;
      }
    }
    return null;
  }

//***************************************************************************//

  /**
   *  Maps the values of the data attributes in the datalist of a
   *  checked out workitem to the input params of the worklet case that will
   *  run as a substitute for the checked out workitem.
   *  The input params for the worklet case are required by the interface's
   *  launchcase() method.
   *  @param coci - the checked out workitem's info object
   *  @return the loaded input params of the new worklet case
   *          (launchCase() requires the input params as a String)
   */
  private String mapItemParamsToWorkletCaseParams(ExternalCase extCase) {

    Element result = new Element(extCase.getDecomopositionID()); // new datalist for worklet
    // extract the name of each worklet input param
    for (int i = 0; i < extCase.dataSize(); i++) {

      CaseDataElement data = extCase.getDataElement(i);
      Element dataElement = new Element(data.getName());
      dataElement.setText(data.getValue());
      result.addContent(dataElement);
    }

    // return the datalist as as string (as required by launchcase)
    return new XMLOutputter().outputString(result);
  }

  //***************************************************************************//
  public void handleCompleteCaseEvent(String caseID, String casedata) {
    if (connected()) {
   /*   Element cdata = null;
      try {
		cdata =  (Element) XMLParser.fromString(casedata);
	} catch (TransformerException e) {
		log.exception(e);
	}*/
      completedCase(caseID, null);
    }
  }

//***************************************************************************//

  protected abstract void completedCase(String caseId, Element data);
//***************************************************************************//


  //***************************************************************************//
  /**
   *  Uploads a worklet specification into the engine
   *  @param workletName - the name of the worklet specification to upoad
   *  @returns true if upload is successful or spec is already loaded
   */
  private boolean uploadSpecification(ExternalCase ec) {

    if (connected()) {
      String fileName = ec.getFileName()/* + ".xml"*/;
      String fullFileName = ec.getFilePath() + fileName;

      if (isUploaded(ec)) {
        log.log("Specification '" + fullFileName
                + "' is already loaded in YAWL Engine");
        return true;
      }

      String wSpec = fileToString(fullFileName); // needs spec as String

      if (wSpec != null) {
        try {
			if (successful(interfaceAClient.uploadSpecification(wSpec,
			   sessionHandle))) {
			  log.log("Successfully uploaded specification: "
			          + fullFileName);
			  return true;
			}
			else {
			  log.log("Unsuccessful specification upload : "
			          + fullFileName);
			  return false;
			}
		} catch (IOException e) {
			log.exception(e);
		}
      }
      else {
        log.log("Rule search found: " + fullFileName +
                ", but there is no worklet of that name in " +
                "the repository, or there was a problem " +
                "opening/reading the worklet specification");
        return false;
      }
    }
    return false;
  }

  //***************************************************************************//

  /** return true if the specification is alreaded loaded in the engine */
  private boolean isUploaded(ExternalCase ec) {

    // refresh list of specifications loaded into the engine
    if (connected()) {
      ArrayList<SpecificationData> loadedSpecs = null;
      try {
        loadedSpecs = (ArrayList<SpecificationData>) _interfaceBClient.getSpecificationList(
            sessionHandle);
      }
      catch (IOException ioe) {
        log.exception(ioe);
      }

      String specification = ec.getStecificationID();
      // check if any loaded specids match the specification of the external case
      for (int i = 0; i < loadedSpecs.size(); i++) {
        SpecificationData spec = (SpecificationData) loadedSpecs.get(i);
        if (specification.equals(spec.getID())){
        	return true;
        }
      }
    }
    return false; // no matches
  }

//***************************************************************************//
  /**
   *  converts the contents of a file to a String
   *  @param fName the name of the file
   *  @return the String representing the file's contents
   */
  private String fileToString(String fName) {
    String fLine;
    StringBuffer result = new StringBuffer();
    try {

      File f = new File(fName);

      if (!f.exists()) {
        return null; // don't go further if no file
      }

      FileReader fread = new FileReader(fName);
      BufferedReader bufread = new BufferedReader(fread);

      fLine = bufread.readLine(); // read first line
      while (fLine != null) {
        result.append(fLine);
        fLine = bufread.readLine(); // read next line
      }
      bufread.close();
      fread.close();
      return result.toString();
    }
    catch (FileNotFoundException fnfe) {
      log.exception(fnfe);
      return null;
    }
    catch (IOException ioe) {
      log.exception(ioe);
      return null;
    }
  }
}
